"""Module for parsing labels returned by kili.labels()."""

from typing import Dict, Generator, Iterable, List, overload

from kili.core.enums import InputType

from ...services.label_data_parsing.label import ParsedLabel


@overload
def parse_labels(
    labels: List[Dict], json_interface: Dict, input_type: InputType
) -> List[ParsedLabel]:
    ...


@overload
def parse_labels(
    labels: Generator[Dict, None, None], json_interface: Dict, input_type: InputType
) -> Generator[ParsedLabel, None, None]:
    ...


def parse_labels(
    labels: Iterable[Dict], json_interface: Dict, input_type: InputType
) -> Iterable[ParsedLabel]:
    """Parse labels returned by kili.labels().

    Args:
        labels: List or generator of labels from kili.labels().
        json_interface: Json interface of the project.
        input_type: Type of assets of the project.

    Returns:
        Parsed labels.
    """
    gen = (
        ParsedLabel(label=label, json_interface=json_interface, input_type=input_type)
        for label in labels
    )
    return gen if isinstance(labels, Generator) else list(gen)



# Feature is still under development and is not yet suitable for use by general users.
"""Classes for json response parsing."""

from typing import Dict

from kili.core.enums import InputType

from .exceptions import JobNotExistingError
from .job_response import JobPayload


class ParsedJobs:
    """Class for label json response parsing."""

    def __init__(self, json_response: Dict, json_interface: Dict, input_type: InputType) -> None:
        """Class for label json response parsing.

        Args:
            json_response: Value of the key "jsonResponse" of a label.
            json_interface: Json interface of the project.
            input_type: Type of assets of the project.
        """
        _ = input_type  # will be used in the future for video parsing
        self._json_data: Dict[str, JobPayload] = {}

        for job_name, job_interface in json_interface["jobs"].items():
            job_response = json_response.get(job_name, {})

            if job_interface["required"] and not job_response:
                raise JobNotExistingError(job_name)

            self._json_data[job_name] = JobPayload(
                job_interface=job_interface,
                job_payload=job_response,
            )

        # check that the job names in the json response are in the json interface
        for job_name in json_response:
            if job_name not in json_interface["jobs"]:
                raise JobNotExistingError(job_name)

    def __getitem__(self, job_name: str) -> JobPayload:
        """Returns the JobPayload object corresponding to the job name."""
        if job_name not in self._json_data:
            raise JobNotExistingError(job_name)
        return self._json_data[job_name]

    def to_dict(self) -> Dict:
        """Returns the parsed json response as a dict."""
        ret = {job_name: job_payload.to_dict() for job_name, job_payload in self._json_data.items()}
        ret = {k: v for k, v in ret.items() if v}  # remove empty json responses
        return ret

    def __repr__(self) -> str:
        """Returns the representation of the object."""
        return repr(self.to_dict())

    def __str__(self) -> str:
        """Returns the string representation of the object."""
        return str(self.to_dict())



# Feature is still under development and is not yet suitable for use by general users.
"""Module for the "categories" key parsing of a job response."""
from typing import Any, Dict, List, Optional

from typeguard import typechecked

from .exceptions import InvalidMutationError


class Category:
    """Class for Category parsing."""

    def __init__(self, category_json: Dict, job_interface: Dict) -> None:
        """Class for Category parsing.

        Args:
            category_json: Value of the key "categories".
            job_interface: Job interface of the job.
        """
        self._json_data = {}
        self._job_interface = job_interface

        self.name = category_json["name"]
        if "confidence" in category_json:
            self.confidence = category_json["confidence"]
        if "children" in category_json:
            self.children = category_json["children"]

    def __str__(self) -> str:
        """Returns the string representation of the category."""
        return str(self._json_data)

    def __repr__(self) -> str:
        """Returns the string representation of the category."""
        return repr(self._json_data)

    def as_dict(self) -> Dict:
        """Returns the parsed category as a dict."""
        return self._json_data

    @property
    def name(self) -> str:
        """Returns the name of the category label."""
        return self._json_data["name"]

    @name.setter
    @typechecked
    def name(self, name: str) -> None:
        """Sets the name of the category label."""
        if name not in self._job_interface["content"]["categories"]:
            raise InvalidMutationError(
                f"Category '{name}' is not in the job interface with categories:"
                f" {self._job_interface['content']['categories'].keys()}"
            )
        self._json_data["name"] = name

    @property
    def confidence(self) -> int:
        """Returns the confidence of the category label."""
        return self._json_data["confidence"]

    @confidence.setter
    @typechecked
    def confidence(self, confidence: int) -> None:
        """Sets the confidence of the category label."""
        if not 0 <= confidence <= 100:
            raise ValueError(f"Confidence must be between 0 and 100, got {confidence}")
        self._json_data["confidence"] = confidence

    @property
    def children(self):
        """Returns the children of the category label."""
        return self._json_data["children"]

    @children.setter
    @typechecked
    def children(self, children: List[Dict[str, Any]]) -> None:
        """Sets the children of the category label."""
        self._json_data["children"] = children


class CategoryList:
    """Class for the categories list parsing."""

    def __init__(self, categories_list: List[Dict[str, Any]], job_interface: Dict) -> None:
        """Class for the categories list parsing.

        Args:
            categories_list: List of dicts representing categories.
            job_interface: Job interface of the job.
        """
        self._categories_list: List[Category] = []
        self._job_interface = job_interface

        for category_dict in categories_list:
            self.add_category(**category_dict)

    def _check_can_append_category(self, category: Category) -> None:
        input_type = self._job_interface["content"]["input"]
        nb_classes = len(self._job_interface["content"]["categories"])
        len_categories = len(self._categories_list)

        if input_type in ("radio", "singleDropdown"):
            if len_categories >= 1:
                raise InvalidMutationError(
                    f"Cannot add more than one category to a {input_type} classification job."
                )

        elif input_type in ("checkbox", "multipleDropdown"):
            if len_categories >= nb_classes:
                raise InvalidMutationError(
                    f"Cannot add more categories than the number of classes to a {input_type}"
                    " classification job."
                )

        else:
            raise ValueError(f"Invalid input type: {input_type}")

        # Check that the name of the category we want to add is not already in the list
        if any(category.name == cat.name for cat in self._categories_list):
            raise InvalidMutationError(
                f"Cannot add a category with name '{category.name}' because a category with the"
                f" same name already exists: {self._categories_list}"
            )

    @typechecked
    def add_category(self, name: str, confidence: Optional[int] = None) -> None:
        """Adds a category object to the CategoryList object."""
        category_dict: Dict[str, object] = {"name": name}
        if confidence is not None:
            category_dict["confidence"] = confidence

        category = Category(
            category_json=category_dict,
            job_interface=self._job_interface,
        )
        self._check_can_append_category(category)
        self._categories_list.append(category)

    @typechecked
    def __getitem__(self, index: int) -> Category:
        """Returns the category object at the given index."""
        return self._categories_list[index]

    def __len__(self) -> int:
        """Returns the number of categories."""
        return len(self._categories_list)

    def __str__(self) -> str:
        """Returns the string representation of the categories list."""
        return str(self.as_list())

    def __repr__(self) -> str:
        """Returns the string representation of the categories list."""
        return repr(self.as_list())

    def as_list(self) -> List[Dict]:
        """Returns the list of categories as a list of dicts."""
        return [category.as_dict() for category in self._categories_list]



# Feature is still under development and is not yet suitable for use by general users.
"""Exceptions for label response module."""


class JobNotExistingError(Exception):
    """Raised when the job does not exist."""

    def __init__(self, job_name: str) -> None:
        """Init."""
        super().__init__(f"Job named '{job_name}' does not exist.")


class AttributeNotCompatibleWithJobError(Exception):
    """Raised when the attribute is not compatible with the job."""

    def __init__(self, attribute_name: str) -> None:
        """Init."""
        super().__init__(f"The attribute '{attribute_name}' is not compatible with the job.")


class InvalidMutationError(Exception):
    """Raised when the mutation is invalid."""



"""Module for label parsing."""

from copy import deepcopy
from typing import Dict

from kili.core.enums import InputType

from .json_response import ParsedJobs


class ParsedLabel(Dict):
    """Class that parses a label."""

    def __init__(self, label: Dict, json_interface: Dict, input_type: InputType) -> None:
        """Class that parses a label.

        The class behaves like a dict but adds the attribute "jobs".

        The original label is not modified.

        Args:
            label: Label to parse.
            json_interface: Json interface of the project.
            input_type: Type of assets of the project.

        !!! Example
            ```python
            label = kili.labels(...)[0]
            parsed_label = ParsedLabel(label, json_interface, input_type)
            print(parsed_label.jobs["JOB_0"].category.name)  # "A"
            ```
        """
        super().__init__(deepcopy(label))

        self.jobs = ParsedJobs(
            json_response=self["jsonResponse"],
            json_interface=json_interface,
            input_type=input_type,
        )

    def to_dict(self) -> Dict:
        """Returns a copy of the parsed label as a dict."""
        ret = {k: deepcopy(v) for k, v in self.items() if k != "jsonResponse"}
        ret["jsonResponse"] = self.jobs.to_dict()
        return ret



"""Module for the "boundingPoly" key parsing of an object detection job response."""

from typing import Dict, List

from typeguard import typechecked
from typing_extensions import Literal


class BoundingPoly:
    """Class for parsing an element of a boundingPoly list."""

    def __init__(self, bounding_poly_json: Dict) -> None:
        """Class for BoundingPoly parsing."""
        self._json_data = bounding_poly_json

    def __str__(self) -> str:
        """Returns the string representation of the boundingPoly object."""
        return str(self._json_data)

    def __repr__(self) -> str:
        """Returns the string representation of the boundingPoly object."""
        return repr(self._json_data)

    def as_dict(self) -> Dict:
        """Returns the boundingPoly object as a dict."""
        return self._json_data

    @property
    def normalized_vertices(self) -> List[Dict[Literal["x", "y"], float]]:
        """Returns the normalized vertices of the bounding polygon."""
        return self._json_data["normalizedVertices"]

    @normalized_vertices.setter
    @typechecked
    def normalized_vertices(
        self, normalized_vertices: List[Dict[Literal["x", "y"], float]]
    ) -> None:
        """Sets the normalized vertices of the bounding polygon."""
        for vertex in normalized_vertices:
            if not 0 <= vertex["x"] <= 1:
                raise ValueError(f"Vertex x coordinate {vertex['x']} is not in [0, 1]")
            if not 0 <= vertex["y"] <= 1:
                raise ValueError(f"Vertex y coordinate {vertex['y']} is not in [0, 1]")
        self._json_data["normalizedVertices"] = normalized_vertices



"""Helpers to create boundingPoly rectangle annotations."""

from typing import Dict, List, Optional, Union

from .point import point_to_normalized_point


def bbox_points_to_normalized_vertices(
    *,
    bottom_left: Dict[str, Union[int, float]],
    bottom_right: Dict[str, Union[int, float]],
    top_right: Dict[str, Union[int, float]],
    top_left: Dict[str, Union[int, float]],
    img_width: Optional[Union[int, float]] = None,
    img_height: Optional[Union[int, float]] = None,
) -> List[Dict[str, float]]:
    # pylint: disable=line-too-long
    """Converts a bounding box defined by its 4 points in pixels coordinates to normalized vertices.

    The output can be used to create a boundingPoly rectangle annotation. See the [documentation](https://docs.kili-technology.com/reference/export-object-entity-detection-and-relation#standard-object-detection) for more details.

    A point is a dict with keys 'x' and 'y', and corresponding values in pixels (int or float).

    Conventions for the input points:

    - The origin is the bottom left corner of the image.
    - x-axis is horizontal and goes from left to right.
    - y-axis is vertical and goes from bottom to top.

    Conventions for the output vertices:

    - The origin is the top left corner of the image.
    - x-axis is horizontal and goes from left to right.
    - y-axis is vertical and goes from top to bottom.

    If the image width and height are provided, the point coordinates will be normalized to [0, 1].
    If not, the method expects the points' coordinates to be already normalized.

    Args:
        bottom_left: Bottom left point of the bounding box.
        bottom_right: Bottom right point of the bounding box.
        top_right: Top right point of the bounding box.
        top_left: Top left point of the bounding box.
        img_width: Width of the image the bounding box is defined in.
        img_height: Height of the image the bounding box is defined in.

    Returns:
        A list of normalized vertices.

    !!! Example
        ```python
        from kili.utils.labels.bbox import bbox_points_to_normalized_vertices

        inputs = {
            bottom_left = {"x": 0, "y": 0},
            bottom_right = {"x": 10, "y": 0},
            top_right = {"x": 10, "y": 10},
            top_left = {"x": 0, "y": 10},
            img_width = 100,
            img_height = 100,
        }
        normalized_vertices = bbox_points_to_normalized_vertices(**inputs)
        json_response = {
            "OBJECT_DETECTION_JOB": {
                "annotations": [
                    {
                        "boundingPoly": [{"normalizedVertices": normalized_vertices}],
                        "categories": [{"name": "CLASS_A"}],
                        "type": "rectangle",
                    }
                ]
            }
        }
        ```
    """
    assert bottom_left["x"] <= bottom_right["x"], "bottom_left.x must be <= bottom_right.x"
    assert top_left["x"] <= top_right["x"], "top_left.x must be <= top_right.x"
    assert bottom_left["y"] <= top_left["y"], "bottom_left.y must be <= top_left.y"
    assert bottom_right["y"] <= top_right["y"], "bottom_right.y must be <= top_right.y"

    if (img_width is None) != (img_height is None):
        raise ValueError("img_width and img_height must be both None or both not None.")

    vertices = [
        point_to_normalized_point(top_left, img_width=img_width, img_height=img_height),
        point_to_normalized_point(bottom_left, img_width=img_width, img_height=img_height),
        point_to_normalized_point(bottom_right, img_width=img_width, img_height=img_height),
        point_to_normalized_point(top_right, img_width=img_width, img_height=img_height),
    ]

    return vertices


def normalized_vertices_to_bbox_points(
    normalized_vertices: List[Dict[str, float]],
    img_width: Optional[Union[int, float]] = None,
    img_height: Optional[Union[int, float]] = None,
) -> Dict[str, Dict[str, float]]:
    # pylint: disable=line-too-long
    """Converts a rectangle normalizedVertices annotation to its 4 points in pixels or in normalized coordinates depending on the image width and height arguments.

    It is the inverse of the method `bbox_points_to_normalized_vertices`.

    A point is a dict with keys 'x' and 'y', and corresponding values in pixels (int or float).

    Conventions for the input vertices:

    - The origin is the top left corner of the image.
    - x-axis is horizontal and goes from left to right.
    - y-axis is vertical and goes from top to bottom.

    Conventions for the output points (top_left, bottom_left, bottom_right, top_right):

    - The origin is the bottom left corner of the image.
    - x-axis is horizontal and goes from left to right.
    - y-axis is vertical and goes from bottom to top.

    If the image width and height are provided, the point coordinates will be scaled to the image size.
    If not, the method will keep the normalized coordinates.

    Args:
        normalized_vertices: A list of normalized vertices.
        img_width: Width of the image the bounding box is defined in.
        img_height: Height of the image the bounding box is defined in.

    Returns:
        A dict with keys 'top_left', 'bottom_left', 'bottom_right', 'top_right', and corresponding points.

    !!! Example
        ```python
        from kili.utils.labels.bbox import normalized_vertices_to_bbox_points

        normalized_vertices = label["jsonResponse"]["OBJECT_DETECTION_JOB"]["annotations"][0]["boundingPoly"][0]["normalizedVertices"]
        img_height, img_width = 1080, 1920
        bbox_points = normalized_vertices_to_bbox_points(normalized_vertices, img_width, img_height)
        ```
    """
    if len(normalized_vertices) != 4:
        raise ValueError("normalized_vertices must have length 4.")

    if (img_width is None) != (img_height is None):
        raise ValueError("img_width and img_height must be both None or both not None.")

    img_height = img_height or 1
    img_width = img_width or 1

    top_left = {}
    bottom_left = {}
    bottom_right = {}
    top_right = {}

    for vertex, point in zip(normalized_vertices, (top_left, bottom_left, bottom_right, top_right)):
        point["x"] = vertex["x"] * img_width
        point["y"] = (1 - vertex["y"]) * img_height

    return {
        "top_left": top_left,
        "bottom_left": bottom_left,
        "bottom_right": bottom_right,
        "top_right": top_right,
    }
